# -*- coding: utf-8 -*- 
# @Time    : 04/03/2024 
# @Author  : nzooherd 
# @File    : video_factory.py 
# @Software: Visual Studio Code

import hashlib
from logging import getLogger
import os
import random
import shutil
import ffmpeg
import pathos

from nzooherd_pyzen.classes import Singleton
from aspect.task import task_logger
from utills.timer_util import second2time


logger = getLogger()


class VideoFactory(metaclass=Singleton):
    """
    VideoFactory
    """

    def __init__(self):
        self.pool = pathos.pools.ProcessPool(4)

    def md5sum(self, file_path: str) -> str:
        """
        Generate video md5
        """
        logger.info(f"Start generate md5 for file: {file_path}")
        with open(file_path, 'rb') as fp:
            file_hash = hashlib.md5()
            while chunk := fp.read(8192 * 1024):
                file_hash.update(chunk)
            logger.info(f"Generate md5 for file: {file_path} is {file_hash.hexdigest()}")
            return file_hash.hexdigest()
        
    
    def preview_video(self, source_file_path: str, target_file_path: str):
        #return self.pool.apipe(self.preview_video_do, source_file_path, target_file_path).get()
        return self.preview_video_do(source_file_path, target_file_path)

    @task_logger
    def preview_video_do(self, source_file_path: str, target_file_path: str):
        """
        Generate video preview
        Our goal is to generate a video trailer within 1 minute. The video will contains about 12 clip, and each clip is 5 second.
        1. Confirm the clips number, randomly choose from [10, 12]
        2. Split the original video the number we choose clips.
        3. Random select some clips to generate a preview video.
        """
        logger.info(f"Start generate preview video: {source_file_path} to {target_file_path}")
        preview_clip_dir = os.path.join(os.path.dirname(source_file_path), os.path.basename(source_file_path).split('.')[0] + "_clip")
        if not os.path.exists(preview_clip_dir):
            os.mkdir(preview_clip_dir)
        if os.path.exists(target_file_path):
            os.remove(target_file_path)

        # TODO tackle rot video 
        video_probe = ffmpeg.probe(source_file_path)

        random.seed()
        clip_number = random.randint(10, 15)
        total_time = int(float(video_probe['format']['duration']))

        clip_edge_timer = {0, total_time}
        while len(clip_edge_timer) < clip_number + 1:
            time = random.randint(0, total_time)
            if time not in clip_edge_timer:
                clip_edge_timer.add(time)
        clip_edge_timer = sorted(clip_edge_timer) 

        clip_timer = []
        remain_time = 60

        for i in range(len(clip_edge_timer) - 1):
            if remain_time <= 0:
                break
            clip_start = clip_edge_timer[i]
            clip_end = clip_edge_timer[i + 1]
            average_reamin_time = remain_time // (clip_number - i)
            if average_reamin_time <= 0:
                break

            duration = random.randint(average_reamin_time - 1, average_reamin_time + 1)
            if duration > clip_end - clip_start:
                duration = clip_end - clip_start
            
            if duration > remain_time:
                duration = remain_time

            remain_time -= duration 
            real_start = random.randint(clip_start, clip_end - duration)
            clip_timer.append((real_start, duration))

        video_clips_name = []

        for i, item in enumerate(clip_timer):
            start, duration = item
            temp_file = os.path.join(preview_clip_dir, f"temp-{i}.mp4")
            logger.info("Start split video clip: %s, %s, %s", temp_file, start, duration)
            stream = ffmpeg.input(source_file_path, ss=second2time(start), to=second2time(start + duration))
            cmd = ffmpeg.output(stream, temp_file).overwrite_output()
            logger.debug(f"ffmpeg {' '.join(cmd.get_args())}")
            cmd.run()
            video_clips_name.append(temp_file)
        
        input_streams = [ffmpeg.input(v) for v in video_clips_name]
        video_streams = [s.video for s in input_streams]
        audio_streams = [s.audio for s in input_streams]
        video_concat = ffmpeg.filter(video_streams, 'concat', n=len(video_streams), v=1, a=0)
        audio_concat = ffmpeg.filter(audio_streams, 'concat', n=len(audio_streams), v=0, a=1)

        output = ffmpeg.output(video_concat, audio_concat, target_file_path)
        ffmpeg.run(output)

        for f in video_clips_name:
            os.remove(f)
        
        shutil.rmtree(preview_clip_dir)

        while True:
            logger.info(f"Check file size: {os.path.getsize(target_file_path)}")
            if os.path.getsize(target_file_path) <= 10 * 1024 * 1024:
                break
            self._compress(target_file_path)

    def thumbnail_photo(self, video_path: str):
        """
        Generate thumbnail photo
        """
        pass

    def _compress(self, source_file_path: str, size_limit: int = 10):
        """
        To enable auto preview in telegram, we need to compress the video to make it smaller than 10MB.

        Parameters
            source_file_path:
            size_limit: unit: MB, default 10
        """
        probe = ffmpeg.probe(source_file_path)
        video_stream = next((stream for stream in probe['streams'] if stream['codec_type'] == 'video'), None)
        width = int(video_stream['width'])
        height = int(video_stream['height'])
        if width > 1280:
            new_width = 1280
            new_height = int(height * 1280 / width)
        else:
            new_width = int(width * 0.8)
            new_height = int(height * 0.8)
        dirname = os.path.dirname(source_file_path)
        basename = os.path.basename(source_file_path)

        input_stream = ffmpeg.input(source_file_path)
        scaled_video = input_stream.video.filter('scale', new_width, new_height)
        temp_target_file = os.path.join(dirname, "temp-" + basename)
        output = ffmpeg.output(scaled_video, input_stream.audio, temp_target_file)
        ffmpeg.run(output)

        os.remove(source_file_path)
        os.rename(temp_target_file, source_file_path)
        # TODO 压缩帧率

    def __del__(self):
        self.pool.close()
